import lgdo
import numpy as np
import pytest

from dspeed.errors import ProcessingChainError
from dspeed.processing_chain import build_processing_chain


def test_waveform_slicing(geds_raw_tbl):
    dsp_config = {
        "outputs": ["wf_blsub"],
        "processors": {
            "wf_blsub": {
                "function": "bl_subtract",
                "module": "dspeed.processors",
                "args": ["waveform[0:100]", "baseline", "wf_blsub"],
                "unit": "ADC",
            },
        },
    }
    proc_chain, _, tbl_out = build_processing_chain(geds_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)

    assert list(tbl_out.keys()) == ["wf_blsub"]
    assert isinstance(tbl_out["wf_blsub"], lgdo.WaveformTable)
    assert tbl_out["wf_blsub"].wf_len == 100


def test_processor_none_arg(geds_raw_tbl):
    dsp_config = {
        "outputs": ["wf_cum"],
        "processors": {
            "wf_cum": {
                "function": "cumsum",
                "module": "numpy",
                "args": ["waveform", 1, None, "wf_cum"],
                "kwargs": {"signature": "(n),(),()->(n)", "types": ["fii->f"]},
                "unit": "ADC",
            }
        },
    }
    proc_chain, _, _ = build_processing_chain(geds_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)

    dsp_config["processors"]["wf_cum"]["args"][2] = "None"
    proc_chain, _, _ = build_processing_chain(geds_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)


def test_processor_kwarg_assignment(geds_raw_tbl):
    dsp_config = {
        "outputs": ["wf_cum"],
        "processors": {
            "wf_cum": {
                "function": "cumsum",
                "module": "numpy",
                "args": ["waveform", "axis=1", "out=wf_cum"],
                "kwargs": {"signature": "(n),()->(n)", "types": ["fi->f"]},
                "unit": "ADC",
            }
        },
    }
    proc_chain, _, _ = build_processing_chain(geds_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)

    dsp_config["processors"]["wf_cum"]["args"][1] = "dtypo=None"
    proc_chain, _, _ = build_processing_chain(geds_raw_tbl, dsp_config)
    with pytest.raises(TypeError):
        proc_chain.execute(0, 1)


def test_processor_dtype_arg(geds_raw_tbl):
    dsp_config = {
        "outputs": ["wf_cum"],
        "processors": {
            "wf_cum": {
                "function": "cumsum",
                "module": "numpy",
                "args": ["waveform", "axis=0", "dtype='int32'", "out=wf_cum"],
                "kwargs": {"signature": "(n),(),()->(n)", "types": ["fiU->i"]},
                "unit": "ADC",
            }
        },
    }
    proc_chain, _, _ = build_processing_chain(geds_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)


def test_scipy_gauss_filter(geds_raw_tbl):
    dsp_config = {
        "outputs": ["wf_gaus"],
        "processors": {
            "wf_gaus": {
                "function": "gaussian_filter1d",
                "module": "scipy.ndimage",
                "args": [
                    "waveform",
                    "0.1*us",
                    "mode='reflect'",
                    "truncate=3",
                    "output=wf_gaus",
                ],
                "kwargs": {"signature": "(n),(),(),()->(n)", "types": ["ffUf->f"]},
                "unit": "ADC",
            }
        },
    }
    proc_chain, _, _ = build_processing_chain(geds_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)


def test_histogram_processor_fixed_width(spms_raw_tbl):
    dsp_config = {
        "outputs": ["hist_weights", "hist_borders"],
        "processors": {
            "hist_weights , hist_borders": {
                "function": "histogram",
                "module": "dspeed.processors.histogram",
                "args": ["waveform", "hist_weights(100)", "hist_borders(101)"],
                "unit": ["none", "ADC"],
            }
        },
    }
    proc_chain, _, _ = build_processing_chain(spms_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)


def test_processor_variable_array_output(spms_raw_tbl):
    dsp_config = {
        "outputs": ["vt_max_out"],
        "processors": {
            "vt_max_out, vt_min_out, n_max_out, n_min_out": {
                "function": "get_multi_local_extrema",
                "module": "dspeed.processors",
                "args": [
                    "waveform",
                    5,
                    5,
                    0,
                    10,
                    0,
                    "vt_max_out(10)",
                    "vt_min_out(10)",
                    "n_max_out",
                    "n_min_out",
                ],
                "unit": "ADC",
            }
        },
    }

    proc_chain, _, _ = build_processing_chain(spms_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)


def test_proc_chain_unit_conversion(spms_raw_tbl):
    dsp_config = {
        "outputs": ["a_unitless", "a_ns", "a_us", "a_ghz"],
        "processors": {
            "a_unitless": {
                "function": "fixed_time_pickoff",
                "module": "dspeed.processors",
                "args": ["waveform", 100, "'n'", "a_unitless"],
            },
            "a_ns": {
                "function": "fixed_time_pickoff",
                "module": "dspeed.processors",
                "args": ["waveform", "1600*ns", "'n'", "a_ns"],
            },
            "a_us": {
                "function": "fixed_time_pickoff",
                "module": "dspeed.processors",
                "args": ["waveform", "1.6*us", "'n'", "a_us"],
            },
            "a_ghz": {  # note this doesn't really make sense, but I want to test if it will convert inverse units
                "function": "fixed_time_pickoff",
                "module": "dspeed.processors",
                "args": ["waveform", "6.25*GHz", "'n'", "a_ghz"],
            },
        },
    }
    proc_chain, _, lh5_out = build_processing_chain(spms_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)
    assert lh5_out["a_unitless"][0] == lh5_out["a_ns"][0]
    assert lh5_out["a_unitless"][0] == lh5_out["a_us"][0]
    assert lh5_out["a_unitless"][0] == lh5_out["a_ghz"][0]


# Test that timing variables can be converted between multiple coordinate
# grids correctly. Also tests slicing with a stride. Pickoff a time from
# a windowed wf and a down-sampled waveform; they should be the same
def test_proc_chain_coordinate_grid(spms_raw_tbl):
    dsp_config = {
        "outputs": ["a_window", "a_downsample", "tp", "tp_window", "tp_downsample"],
        "processors": {
            "a_window": {
                "function": "fixed_time_pickoff",
                "module": "dspeed.processors",
                "args": [
                    "waveform[2625:4025]",
                    "51.2*us + waveform.offset",
                    "'i'",
                    "a_window",
                ],
                "unit": ["ADC"],
            },
            "a_downsample": {
                "function": "fixed_time_pickoff",
                "module": "dspeed.processors",
                "args": [
                    "waveform[0:8000:8]",
                    "51.2*us + waveform.offset",
                    "'i'",
                    "a_downsample",
                ],
                "unit": ["ADC"],
            },
            "tp": {
                "function": "time_point_thresh",
                "module": "dspeed.processors",
                "args": ["waveform", "a_window", "52.48*us+waveform.offset", 0, "tp"],
                "unit": "ns",
            },
            "tp_window": {
                "function": "time_point_thresh",
                "module": "dspeed.processors",
                "args": [
                    "waveform[2625:4025]",
                    "a_window",
                    "52.48*us+waveform.offset",
                    0,
                    "tp_window",
                ],
                "unit": "ns",
            },
            "tp_downsample": {
                "function": "time_point_thresh",
                "module": "dspeed.processors",
                "args": [
                    "waveform[0:8000:8]",
                    "a_window",
                    "52.48*us+waveform.offset",
                    0,
                    "tp_downsample",
                ],
                "unit": "ns",
            },
        },
    }

    proc_chain, _, lh5_out = build_processing_chain(spms_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)
    assert lh5_out["a_window"][0] == lh5_out["a_downsample"][0]
    assert lh5_out["tp_window"][0] == lh5_out["tp"][0]
    assert -128 < lh5_out["tp_downsample"][0] - lh5_out["tp"][0] < 128


def test_proc_chain_round(spms_raw_tbl):
    dsp_config = {
        "outputs": ["w_round", "w_floor", "w_ceil", "w_trunc"],
        "processors": {
            "w_round": "round(waveform, 4)",
            "w_floor": "floor(waveform, 4)",
            "w_ceil": "ceil(waveform, 4)",
            "w_trunc": "trunc(waveform, 4)",
        },
    }

    proc_chain, _, lh5_out = build_processing_chain(spms_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)
    wf = spms_raw_tbl["waveform"].values[0]
    assert np.all(np.rint(wf / 4) * 4 == lh5_out["w_round"].values[0])
    assert np.all(np.floor(wf / 4) * 4 == lh5_out["w_floor"].values[0])
    assert np.all(np.ceil(wf / 4) * 4 == lh5_out["w_ceil"].values[0])
    assert np.all(np.trunc(wf / 4) * 4 == lh5_out["w_trunc"].values[0])


def test_proc_chain_where(spms_raw_tbl):
    # test with variable and const
    dsp_config = {
        "outputs": [
            "tp_min",
            "tp_max",
            "wf_min",
            "wf_max",
            "test1",
            "test2",
            "test3",
            "test4",
            "test5",
            "test6",
        ],
        "processors": {
            "tp_min, tp_max, wf_min, wf_max": {
                "function": "min_max",
                "module": "dspeed.processors",
                "args": ["waveform", "tp_min", "tp_max", "wf_min", "wf_max"],
                "unit": ["ns", "ns", "ADC", "ADC"],
            },
            "test1": "where(waveform<0, 0, waveform)",
            "test2": "where(waveform<0, waveform, 0)",
            "test3": "where(eventnumber==0, tp_min, 1*ns)",
            "test4": "where(eventnumber==0, tp_min, 1*us)",
            "test5": "where(eventnumber==0, 1*ns, tp_min)",
            "test6": "where(eventnumber==0, 1*us, tp_min)",
            "test7": "where(eventnumber==0, tp_min, wf_min)",
        },
    }

    proc_chain, _, lh5_out = build_processing_chain(spms_raw_tbl, dsp_config)
    proc_chain.execute(0, 2)
    wf = spms_raw_tbl["waveform"].values[0]
    assert np.all(np.where(wf < 0, 0, wf) == lh5_out["test1"].values[0])
    assert np.all(np.where(wf < 0, wf, 0) == lh5_out["test2"].values[0])
    tp_min = lh5_out["tp_min"].nda
    assert lh5_out["test3"].attrs["units"] == "nanosecond"
    assert lh5_out["test3"].nda[0] == tp_min[0] and lh5_out["test3"].nda[1] == 1
    assert lh5_out["test4"].attrs["units"] == "nanosecond"
    assert lh5_out["test4"].nda[0] == tp_min[0] and lh5_out["test4"].nda[1] == 1000
    assert lh5_out["test5"].attrs["units"] == "nanosecond"
    assert lh5_out["test5"].nda[0] == 1 and lh5_out["test5"].nda[1] == tp_min[1]
    assert lh5_out["test6"].attrs["units"] == "nanosecond"
    assert lh5_out["test6"].nda[0] == 1000 and lh5_out["test6"].nda[1] == tp_min[1]
    with pytest.raises(ProcessingChainError):
        proc_chain, _, lh5_out = build_processing_chain(
            spms_raw_tbl, dsp_config, outputs=["test7"]
        )

    # test with variable and variable
    dsp_config = {
        "processors": {
            "w_downsample": "waveform[::2]",
            "w_win1": "waveform[:len(waveform)//2]",
            "w_win2": "waveform[len(waveform)//2:]",
            "tp_min, tp_max, wf_min, wf_max": {
                "function": "min_max",
                "module": "dspeed.processors",
                "args": ["waveform", "tp_min", "tp_max", "wf_min", "wf_max"],
                "unit": ["ns", "ns", "ADC", "ADC"],
            },
            "delta_t": "tp_max - tp_min",
            "test1": "where(eventnumber==0, w_downsample, w_win1)",  # fail due to different periods
            "test2": "where(eventnumber==0, w_win1, w_win2)",  # different offsets
            "test3": "where(eventnumber==0, tp_max, delta_t)",  # fail to different is_coord
            "test4": "where(eventnumber==0, tp_min, tp_max)",
        }
    }

    with pytest.raises(ProcessingChainError):
        proc_chain, _, lh5_out = build_processing_chain(
            spms_raw_tbl, dsp_config, outputs=["test1"]
        )

    proc_chain, _, lh5_out = build_processing_chain(
        spms_raw_tbl, dsp_config, outputs=["test2", "w_win1", "w_win2"]
    )
    proc_chain.execute(0, 2)
    assert (
        np.all(lh5_out["test2"].values[0] == lh5_out["w_win1"].values[0])
        and lh5_out["test2"].t0[0] == lh5_out["w_win1"].t0[0]
    )
    assert (
        np.all(lh5_out["test2"].values[1] == lh5_out["w_win2"].values[1])
        and lh5_out["test2"].t0[1] == lh5_out["w_win2"].t0[1]
    )

    with pytest.raises(ProcessingChainError):
        proc_chain, _, lh5_out = build_processing_chain(
            spms_raw_tbl, dsp_config, outputs=["test3"]
        )

    proc_chain, _, lh5_out = build_processing_chain(
        spms_raw_tbl, dsp_config, outputs=["test4", "tp_min", "tp_max"]
    )
    proc_chain.execute(0, 2)
    assert lh5_out["test4"].attrs["units"] == "nanosecond"
    assert np.all(lh5_out["test4"].nda[0] == lh5_out["tp_min"].nda[0])
    assert np.all(lh5_out["test4"].nda[1] == lh5_out["tp_max"].nda[1])

    # test with choosing between constant values
    dsp_config = {
        "processors": {
            "test1": "where(eventnumber==0, 10*ns, 1*us, dtype='f')",
            "test2": "where(eventnumber==0, 10*ns, 1000, dtype='f')",
            "test3": "where(eventnumber==0, 1000, 10*ns, dtype='f')",
            "test4": "where(eventnumber==0, 10, 1000, dtype='f')",
            "test5": "where(eventnumber==0, 10*ns, 10*m, dtype='f')",
        }
    }
    proc_chain, _, lh5_out = build_processing_chain(
        spms_raw_tbl, dsp_config, outputs=["test1", "test2", "test3", "test4"]
    )
    proc_chain.execute(0, 2)
    assert lh5_out["test1"].attrs["units"] == "nanosecond"
    assert lh5_out["test1"].nda[0] == 10 and lh5_out["test1"].nda[1] == 1000
    assert lh5_out["test2"].attrs["units"] == "nanosecond"
    assert lh5_out["test2"].nda[0] == 10 and lh5_out["test2"].nda[1] == 1000
    assert lh5_out["test3"].attrs["units"] == "nanosecond"
    assert lh5_out["test3"].nda[0] == 1000 and lh5_out["test3"].nda[1] == 10
    assert lh5_out["test4"].nda[0] == 10 and lh5_out["test4"].nda[1] == 1000
    with pytest.raises(ProcessingChainError):
        proc_chain, _, lh5_out = build_processing_chain(
            spms_raw_tbl, dsp_config, outputs=["test5"]
        )

    # test a if b else c form
    dsp_config = {
        "outputs": ["test"],
        "processors": {
            "test": "0 if waveform<0 else waveform",
        },
    }

    proc_chain, _, lh5_out = build_processing_chain(spms_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)
    wf = spms_raw_tbl["waveform"].values[0]
    assert np.all(np.where(wf < 0, 0, wf) == lh5_out["test"].values[0])


def test_proc_chain_isnan():
    dsp_config = {
        "outputs": ["test_nan", "test_finite"],
        "processors": {
            "test_nan": "isnan(input)",
            "test_finite": "isfinite(input)",
        },
    }

    tb = lgdo.Table(
        {"input": lgdo.Array(np.array([1.0, 0.0, np.inf, -np.inf, np.nan]))}
    )
    proc_chain, _, lh5_out = build_processing_chain(tb, dsp_config)
    proc_chain.execute()
    assert np.all(
        np.array([False, False, False, False, True]) == lh5_out["test_nan"].nda
    )
    assert np.all(
        np.array([True, True, False, False, False]) == lh5_out["test_finite"].nda
    )


def test_proc_chain_as_type(spms_raw_tbl):
    dsp_config = {
        "outputs": ["waveform_32"],
        "processors": {"waveform_32": "astype(waveform, 'float32')"},
    }

    proc_chain, _, lh5_out = build_processing_chain(spms_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)
    assert np.all(
        spms_raw_tbl["waveform"].values[0] == lh5_out["waveform_32"].values[0]
    )


def test_output_types(spms_raw_tbl):
    dsp_config = {
        "outputs": ["wf_out", "vov_max_out", "n_max_out", "aoa_out"],
        "processors": {
            "wf_out": "-waveform",
            "aoa_out": "n_max_out + [1, 3, 5, 7, 9]",
            "vov_max_out, vov_min_out, n_max_out, n_min_out": {
                "function": "get_multi_local_extrema",
                "module": "dspeed.processors.get_multi_local_extrema",
                "args": [
                    "waveform",
                    5,
                    0.1,
                    1,
                    10,
                    0,
                    "vov_max_out(20, vector_len=n_max_out)",
                    "vov_min_out(20, vector_len=n_min_out)",
                    "n_max_out",
                    "n_min_out",
                ],
                "unit": ["ns", "ns", "none", "none"],
            },
        },
    }

    proc_chain, _, lh5_out = build_processing_chain(spms_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)
    assert isinstance(lh5_out["n_max_out"], lgdo.Array)
    assert isinstance(lh5_out["wf_out"], lgdo.WaveformTable)
    assert isinstance(lh5_out["aoa_out"], lgdo.ArrayOfEqualSizedArrays)
    assert isinstance(lh5_out["vov_max_out"], lgdo.VectorOfVectors)


def test_output_attrs(geds_raw_tbl):
    dsp_config = {
        "outputs": ["wf_blsub"],
        "processors": {
            "wf_blsub": {
                "function": "bl_subtract",
                "module": "dspeed.processors",
                "args": ["waveform[0:100]", "baseline", "wf_blsub"],
                "unit": "ADC",
                "lh5_attrs": {"test_attr": "This is a test"},
            }
        },
    }
    proc_chain, _, lh5_out = build_processing_chain(geds_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)
    assert lh5_out["wf_blsub"].attrs["test_attr"] == "This is a test"


def test_database_params(geds_raw_tbl):
    dsp_config = {
        "outputs": ["test"],
        "processors": {
            "dbabc": "waveform[0]*0",
            "redberry": "dbabc+1",
            "test": {
                "function": "db.a + dbabc + redberry + db.b*db.c",
                "defaults": {"db.a": 1, "db.b": 2, "db.c": 3},
            },
        },
    }

    proc_chain, _, lh5_out = build_processing_chain(geds_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)
    assert lh5_out["test"][0] == 8

    proc_chain, _, lh5_out = build_processing_chain(
        geds_raw_tbl, dsp_config, db_dict={"a": 2, "c": 0}
    )
    proc_chain.execute(0, 1)
    assert lh5_out["test"][0] == 3



from __future__ import annotations

import numpy as np
from numba import guvectorize

from ..errors import DSPFatal
from ..utils import numba_defaults_kwargs as nb_kwargs
from .convolutions import convolve_damped_oscillator


@guvectorize(
    [
        "void(float32[:], float64, float64, float64, float64, float32[:])",
        "void(float64[:], float64, float64, float64, float64, float64[:])",
    ],
    "(n),(),(),(),()->(n)",
    **nb_kwargs(forceobj=True),
)
def inject_damped_oscillation(
    w_in: np.ndarray,
    tau: float,
    omega: float,
    phase: float,
    frac: float,
    w_out: np.ndarray,
) -> None:
    """
    Inject a damped oscillation component/pole into the electronics response

    Parameters
    ----------
    w_in
        the input waveform.
    tau
        time constant of decay
    omega:
        angular frequency of oscillation
    phase:
        phase shift of oscillation
    frac
        fraction of amplitude in injected pole
    w_out
        output waveform after injecting decay component
    """
    if np.isnan(w_in).any():
        return

    if not 0 <= frac <= 1:
        raise DSPFatal("frac must be between zero and one.")

    curr = np.zeros_like(w_out)
    curr[0] = w_in[0]
    curr[1:] = w_in[1:] - w_in[:-1]

    convolve_damped_oscillator(curr, tau, omega, phase, w_out)
    w_out *= frac
    w_out += w_in



r"""
The *dspeed* signal processing framework is responsible for running a variety
of discrete signal processors on data.
"""

from ._version import version as __version__
from .build_dsp import build_dsp
from .processing_chain import ProcessingChain, build_processing_chain

__all__ = ["build_dsp", "ProcessingChain", "build_processing_chain", "__version__"]



import numpy as np
import pytest

from dspeed.errors import DSPFatal
from dspeed.processors import double_pole_zero, pole_zero


def test_pole_zero(compare_numba_vs_python):
    """
    Test that the pole-zero filter can correct an RC decay pulse into a step function
    """

    # Create a single exponential pulse to pole-zero correct
    tau = 30000
    ts = np.arange(0, 8192, dtype=np.float64)
    amplitude = 17500
    pulse_in = np.zeros(len(ts) + 20, dtype=np.float32)
    pulse_in[20:] = amplitude * np.exp(-ts / tau, dtype=np.float32)

    # ensure that if there is a nan in w_in, all nans are outputted
    w_in = np.ones(pulse_in.size)
    w_in[4] = np.nan

    assert np.all(np.isnan(compare_numba_vs_python(pole_zero, w_in, tau)))

    # ensure that a valid input gives the expected output when comparing the pole-zero correction with the pole-zero processor
    step = np.full(len(ts), amplitude)
    w_out_expected = np.insert(step, 0, np.zeros(20))

    # Check that it works at float32 precision
    pulse_in = np.zeros(len(ts) + 20, dtype=np.float32)
    pulse_in[20:] = amplitude * np.exp(-ts / tau)
    result = compare_numba_vs_python(pole_zero, pulse_in, tau)
    assert result.dtype == np.float32
    assert np.allclose(result, w_out_expected, rtol=1e-06)

    # Check that it works at float64 precision
    pulse_in = np.zeros(len(ts) + 20, dtype=np.float64)
    pulse_in[20:] = amplitude * np.exp(-ts / tau)
    result = compare_numba_vs_python(pole_zero, pulse_in, tau)
    assert result.dtype == np.float64
    assert np.allclose(result, w_out_expected, rtol=1e-07)


def test_double_pole_zero(compare_numba_vs_python):
    """
    Test that the double pole-zero filter can correct the sum of two RC decays into a step function
    """

    # Create a double exponential pulse to double-pole-zero correct
    wf_len = 8192
    tp0 = 20
    amplitude = 17500
    tau1 = 1000
    tau2 = 30000
    frac = 0.98
    ts = np.arange(0, wf_len - tp0, dtype=np.float64)
    ys = amplitude * (1 - frac) * np.exp(-ts / tau1) + amplitude * frac * np.exp(
        -ts / tau2
    )

    # ensure the DSPFatal is raised if the waveform is too short
    pulse_out = np.zeros(2, dtype=np.float64)
    with pytest.raises(DSPFatal):
        double_pole_zero(np.ones(2), tau1, tau2, frac, pulse_out)

    # ensure that if there is a nan in w_in, all nans are outputted
    w_in = np.ones(wf_len)
    w_in[4] = np.nan

    assert np.all(
        np.isnan(compare_numba_vs_python(double_pole_zero, w_in, tau1, tau2, frac))
    )

    # ensure that a valid input gives the expected output when comparing the pole-zero correction with the pole-zero processor
    w_out_expected = np.full(wf_len, amplitude)
    w_out_expected[:tp0] = 0

    pulse_in = np.zeros(wf_len, dtype=np.float64)
    pulse_in[tp0:] = ys
    result = compare_numba_vs_python(double_pole_zero, pulse_in, tau1, tau2, frac)
    assert result.dtype == np.float64
    assert np.allclose(result, w_out_expected, rtol=1e-7)

    # Make sure that the processor also works for float32 precision
    pulse_in = np.zeros(wf_len, dtype=np.float32)
    pulse_in[tp0:] = ys
    result = compare_numba_vs_python(double_pole_zero, pulse_in, tau1, tau2, frac)
    assert result.dtype == np.float32
    assert np.allclose(result, w_out_expected, rtol=1e-6)



from __future__ import annotations

import numpy as np
from numba import vectorize

from ..utils import numba_defaults_kwargs as nb_kwargs


@vectorize(
    [
        f"{t}({t}, {t})"
        for t in ["u1", "u2", "u4", "u8", "i1", "i2", "i4", "i8", "f4", "f8"]
    ],
    **nb_kwargs,
)
def round_to_nearest(val: np.ndarray, to_nearest: int | float) -> None:
    """
    Round value to nearest multiple of to_nearest.

    Parameters
    ----------
    val
        value to be rounded
    to_nearest
        round to multiple of this
    out
        rounded value

    JSON Configuration Example
    --------------------------

    Note: this processor is aliased using `round` in ProcessingChain.
    The following two examples are equivalent.

    .. code-block :: json

        "t_rounded": {
            "function": "round_to_nearest",
            "module": "dspeed.processors",
            "args": ["t_in", "1*us", "t_rounded"]
            "unit": ["ns"]
        },
        "t_rounded": "round(t_in, 1*us)"
    """

    if np.isnan(val):
        return np.nan
    else:
        return to_nearest * round(val / to_nearest)


@vectorize(
    [
        f"{t}({t}, {t})"
        for t in ["u1", "u2", "u4", "u8", "i1", "i2", "i4", "i8", "f4", "f8"]
    ],
    **nb_kwargs,
)
def floor_to_nearest(val: np.ndarray, to_nearest: int | float) -> None:
    """
    Return largest multiple of to_nearest that is <= val

    Parameters
    ----------
    val
        value to be floored
    to_nearest
        floor to multiple of this
    out
        floored value

    JSON Configuration Example
    --------------------------

    Note: this processor is aliased using `floor` in ProcessingChain.
    The following two examples are equivalent.

    .. code-block :: json

        "t_floor": {
            "function": "floor_to_nearest",
            "module": "dspeed.processors",
            "args": ["t_in", "1*us", "t_floor"]
            "unit": ["ns"]
        },
        "t_floor": "floor(t_in, 1*us)"
    """

    if np.isnan(val):
        return np.nan
    else:
        return to_nearest * np.floor(val / to_nearest)


@vectorize(
    [
        f"{t}({t}, {t})"
        for t in ["u1", "u2", "u4", "u8", "i1", "i2", "i4", "i8", "f4", "f8"]
    ],
    **nb_kwargs,
)
def ceil_to_nearest(val: np.ndarray, to_nearest: int | float) -> None:
    """
    Return smallest multiple of to_nearest that is >= val.

    Parameters
    ----------
    val
        value to be ceiled
    to_nearest
        ceil to multiple of this
    out
        ceiled value

    JSON Configuration Example
    --------------------------

    Note: this processor is aliased using `ceil` in ProcessingChain.
    The following two examples are equivalent.

    .. code-block :: json

        "t_ceil": {
            "function": "ceil_to_nearest",
            "module": "dspeed.processors",
            "args": ["t_in", "1*us", "t_ceil"]
            "unit": ["ns"]
        },
        "t_ceil": "ceil(t_in, 1*us)"
    """

    if np.isnan(val):
        return np.nan
    else:
        return to_nearest * np.ceil(val / to_nearest)


@vectorize(
    [
        f"{t}({t}, {t})"
        for t in ["u1", "u2", "u4", "u8", "i1", "i2", "i4", "i8", "f4", "f8"]
    ],
    **nb_kwargs,
)
def trunc_to_nearest(val: np.ndarray, to_nearest: int | float) -> None:
    """
    Return multiple of to_nearest that is closest to val, towards zero

    Parameters
    ----------
    val
        value to be truncated
    to_nearest
        truncate to multiple of this
    out
        truncated value

    JSON Configuration Example
    --------------------------

    Note: this processor is aliased using `ceil` in ProcessingChain.
    The following two examples are equivalent.

    .. code-block :: json

        "t_trunc": {
            "function": "trunc_to_nearest",
            "module": "dspeed.processors",
            "args": ["t_in", "1*us", "t_trunc"]
            "unit": ["ns"]
        },
        "t_trunc": "trunc(t_in, 1*us)"
    """

    if np.isnan(val):
        return np.nan
    else:
        return to_nearest * np.trunc(val / to_nearest)



import numpy as np

from dspeed.processors import (
    ceil_to_nearest,
    floor_to_nearest,
    round_to_nearest,
    trunc_to_nearest,
)


def test_round_to_nearest(compare_numba_vs_python):
    w_in = np.array([-2.0, -1.5, -1.0, -0.75, -0.1, 0.0, 0.1, 0.75, 1.0, 1.5, 2.0])

    assert np.all(
        compare_numba_vs_python(round_to_nearest, w_in, 1.5)
        == np.array([-1.5, -1.5, -1.5, 0.0, 0.0, 0.0, 0.0, 0.0, 1.5, 1.5, 1.5])
    )

    assert np.all(
        compare_numba_vs_python(floor_to_nearest, w_in, 1.5)
        == np.array([-3.0, -1.5, -1.5, -1.5, -1.5, 0.0, 0.0, 0.0, 0.0, 1.5, 1.5])
    )

    assert np.all(
        compare_numba_vs_python(ceil_to_nearest, w_in, 1.5)
        == np.array([-1.5, -1.5, 0.0, 0.0, 0.0, 0.0, 1.5, 1.5, 1.5, 1.5, 3.0])
    )

    assert np.all(
        compare_numba_vs_python(trunc_to_nearest, w_in, 1.5)
        == np.array([-1.5, -1.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.5, 1.5])
    )

    w_in = np.nan
    assert np.isnan(compare_numba_vs_python(round_to_nearest, w_in, 1.5))
    assert np.isnan(compare_numba_vs_python(floor_to_nearest, w_in, 1.5))
    assert np.isnan(compare_numba_vs_python(ceil_to_nearest, w_in, 1.5))
    assert np.isnan(compare_numba_vs_python(trunc_to_nearest, w_in, 1.5))



import copy
import os
import shutil
import uuid
from getpass import getuser
from inspect import unwrap
from pathlib import Path
from tempfile import gettempdir

import numpy as np
import pytest
from legendtestdata import LegendTestData
from lgdo.lh5 import read

from dspeed.utils import GUFuncWrapper

config_dir = Path(__file__).parent / "dsp" / "configs"
_tmptestdir = os.path.join(
    gettempdir(), f"dspeed-tests-{getuser()}-{str(uuid.uuid4())}"
)


@pytest.fixture(scope="session")
def tmptestdir():
    os.mkdir(_tmptestdir)
    return _tmptestdir


def pytest_sessionfinish(session, exitstatus):
    if exitstatus == 0:
        shutil.rmtree(_tmptestdir)


@pytest.fixture(scope="session")
def lgnd_test_data():
    ldata = LegendTestData()
    ldata.checkout("69aaeee")
    return ldata


@pytest.fixture(scope="session")
def geds_raw_tbl(lgnd_test_data):
    obj = read(
        "/geds/raw",
        lgnd_test_data.get_path("lh5/LDQTA_r117_20200110T105115Z_cal_geds_raw.lh5"),
        n_rows=10,
    )
    return obj


@pytest.fixture(scope="session")
def spms_raw_tbl(lgnd_test_data):
    obj = read(
        "/ch0/raw",
        lgnd_test_data.get_path("lh5/L200-comm-20211130-phy-spms.lh5"),
        n_rows=10,
    )
    return obj


@pytest.fixture(scope="session")
def compare_numba_vs_python():
    def numba_vs_python(func, *inputs, signature_override=None):
        """
        Function for testing that the numba and python versions of a
        function are equal.

        Parameters
        ----------
        func
            The Numba-wrapped function to be tested
        *inputs
            The various inputs to be passed to the function to be
            tested.
        signature_override
            Function shape signature to use in the special case that
            outputs are sent as inputs (due to need for new output shape)
            but at least one output is a scalar. In this case, scalars
            will be passed as read-only. To remedy this, rewrite the
            signature with a `->`. For example, say we have a function
            with signature `(n),(m),()` with in output and two outputs.
            You would pass `(n)->(m),()` for the signature override.
            I wish there was another way, but this is how it has to be.

        Returns
        -------
        func_output
            The output of the function to be used in a unit test.

        """

        if signature_override:
            sig = signature_override
            copy_out = False
        elif func.signature:
            sig = func.signature
            copy_out = False
        else:
            # function was implemented with @vectorize
            sig = ",".join(["()"] * func.nin) + "->" + ",".join(["()"] * func.nout)
            copy_out = True

        if len(inputs) == func.nargs:
            # outputs passed as inputs; required when size of output
            # differs from size of input. In this case apply in-place
            # and check/return all args

            # numba outputs
            func(*inputs)
            outputs_numba = [copy.deepcopy(arg) for arg in inputs]

            # unwrapped python outputs
            func_unwrapped = GUFuncWrapper(
                unwrap(func),
                sig,
                "".join([np.array(ar).dtype.char for ar in outputs_numba]),
                copy_out=False,
            )
            func_unwrapped(*inputs)
            outputs_python = [copy.deepcopy(arg) for arg in inputs]

        elif len(inputs) == func.nin:
            # outputs not passed as inputs. Copy and return only outputs

            # numba outputs
            outputs_numba = func(*inputs)
            if func.nout == 1:
                outputs_numba = [outputs_numba]

            # unwrapped python outputs
            types = (
                "".join([np.array(ar).dtype.char for ar in inputs])
                + "->"
                + "".join([np.array(ar).dtype.char for ar in outputs_numba])
            )
            func_unwrapped = GUFuncWrapper(unwrap(func), sig, types, copy_out=copy_out)

            # now outputs must be passed as args. Scalars must be
            # converted to rank 1 arrays
            outputs_python = []
            scalars = []
            for i, ar in enumerate(outputs_numba):
                if len(ar.shape) == 0:
                    outputs_python.append(np.zeros_like(ar).reshape(1))
                    scalars.append(i)
                else:
                    outputs_python.append(np.zeros_like(ar))

            func_unwrapped(*inputs, *outputs_python)

            # convert rank 1 arrays back to scalars
            for i in scalars:
                outputs_python[i] = outputs_python[i][0]
        else:
            raise ValueError("Must pass either all inputs or all inputs and outputs")

        # assert that numba and python are the same up to floating point
        # precision, setting nans to be equal
        assert all(
            np.allclose(o_nb, o_py, equal_nan=True)
            for o_nb, o_py in zip(outputs_numba, outputs_python)
        )

        # return value for comparison with expected solution
        return outputs_numba if len(outputs_numba) > 1 else outputs_numba[0]

    return numba_vs_python



from __future__ import annotations

import numpy as np

from ..utils import GUFuncWrapper


def tf_model(filepath: str) -> GUFuncWrapper:
    """
    Initializer to load a tensorflow model to use in a ProcessingChain

    Parameters
    ----------
    filepath
        name of keras file containing model


    JSON Configuration Example
    --------------------------
    .. code-block :: json

        "classifier":{
            "function": "dspeed.processors.tf_model",
            "args": ["input", "output"],
            "init_args": ["'model.keras'"]
        }
    """
    import tensorflow as tf

    try:
        model = tf.keras.models.load_model(filepath)

        # TODO: form the signature from the model inputs and outputs

        return GUFuncWrapper(
            lambda i: model(i).numpy().squeeze(),
            name=filepath,
            signature="(n)->()",
            types=["f->f", "d->d"],
            vectorized=True,
        )
    except OSError:
        return GUFuncWrapper(
            lambda i: np.nan,
            name="tfmodel_null",
            signature="(n)->()",
            types=["f->f", "d->d"],
        )



from __future__ import annotations

import numpy as np
from numba import guvectorize
from pywt import downcoef

from ..errors import DSPFatal
from ..utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    [
        "void(float32[:], int32, char, char, float32[:])",
        "void(float64[:], int64, char, char, float64[:])",
    ],
    "(n),(),(),(),(m)",
    **nb_kwargs(
        forceobj=True,
    ),
)
def discrete_wavelet_transform(
    w_in: np.ndarray, level: int, wave_type: int, coeff: int, w_out: np.ndarray
) -> None:
    """
    Apply a discrete wavelet transform to the waveform and return only
    the detailed or approximate coefficients.


    Parameters
    ----------

    w_in
       The input waveform
    level
       The level of decompositions to be performed ``(1, 2, ...)``
    wave_type
       The wavelet type for discrete convolution ``('h' = 'haar', 'd' = 'db1')``.
    coeff
       The coefficients to be saved ``('a', 'd')``
    w_out
       The approximate coefficients. The dimension of this array can be calculated
       by ``out_dim = len(w_in)/(filter_length^level)``, where ``filter_length``
       can be obtained via ``pywt.Wavelet(wave_type).dec_len``


    JSON Configuration Example
    --------------------------
    .. code-block :: json

        "dwt_haar":{
            "function": "discrete_wavelet_transform",
            "module": "dspeed.processors",
            "args": ["wf_blsub", 5, "'h'", "'a'", "dwt_haar(256, 'f')"],
            "unit": "ADC",
            "prereqs": ["wf_blsub"],
        }
    """

    w_out[:] = np.nan

    if level <= 0:
        raise DSPFatal("The level must be a positive integer")

    if np.isnan(w_in).any():
        return

    coeff = chr(coeff)

    if chr(wave_type) == "h":
        wave_type = "haar"

    elif chr(wave_type) == "d":
        wave_type = "db1"

    w_out[:] = downcoef(coeff, w_in, wave_type, level=level)



from __future__ import annotations

import numpy as np
from numba import guvectorize

from dspeed.errors import DSPFatal
from dspeed.utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    [
        "void(float32[:], float32, float32, float32, float32[:])",
        "void(float64[:], float64, float64, float64, float64[:])",
    ],
    "(n),(),(),(),(m)",
    **nb_kwargs,
)
def wf_alignment(
    w_in: np.ndarray, centroid: int, shift: int, size: int, w_out: np.ndarray
) -> None:
    """Align waveform.

    Note
    ----
    This processor align the input waveform by setting the centroid position at the center of the output waveform.

    Parameters
    ----------
    w_in
        the input waveform.
    centroid
        centroid.
    shift
        shift.
    size
        size of output waveform.
    w_out
        aligned waveform.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "wf_align": {
          "function": "wf_alignment",
          "module": "dspeed.processors",
          "args": ["waveform", "centroid", "shift", "size", "wf_align"],
          "unit": "ADC"
        }
    """

    w_out[:] = np.nan

    if np.isnan(w_in).any():
        return

    if np.isnan(centroid):
        raise DSPFatal("centroid is nan")

    if np.isnan(shift):
        raise DSPFatal("shift is nan")
    if shift < 0:
        raise DSPFatal("shift must be positive")
    if shift > len(w_in):
        raise DSPFatal("shift must be shorter than input waveform size")

    if np.isnan(size):
        raise DSPFatal("size is nan")
    if size <= 0:
        raise DSPFatal("size must be positive")
    if size > len(w_in):
        raise DSPFatal("size must be shorter than input waveform size")

    if (centroid >= size / 2) and (centroid < len(w_in) - size / 2):
        w_out[:] = w_in[int(centroid - size / 2) : int(centroid + size / 2)]
    elif (centroid > size / 2 - shift) and (centroid < size / 2):
        ss = int((size + 1) / 2 - centroid)
        w_out[:ss] = w_in[0]
        w_out[ss:] = w_in[: int(centroid + size / 2)]
    else:
        w_out[:] = w_in[:size]



from __future__ import annotations

import numpy as np
from numba import guvectorize

from ..errors import DSPFatal
from ..utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    [
        "void(float32[:], float32, float32, float32, float32[:])",
        "void(float64[:], float64, float64, float64, float64[:])",
    ],
    "(n),(),(),()->()",
    **nb_kwargs,
)
def time_point_thresh(
    w_in: np.ndarray, a_threshold: float, t_start: int, walk_forward: int, t_out: float
) -> None:
    """Find the index where the waveform value crosses the threshold, walking
    either forward or backward from the starting index.

    Parameters
    ----------
    w_in
        the input waveform.
    a_threshold
        the threshold value.
    t_start
        the starting index.
    walk_forward
        the backward (``0``) or forward (``1``) search direction.
    t_out
        the index where the waveform value crosses the threshold.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "tp_0": {
            "function": "time_point_thresh",
            "module": "dspeed.processors",
            "args": ["wf_atrap", "bl_std", "tp_start", 0, "tp_0"],
            "unit": "ns"
        }
    """
    t_out[0] = np.nan

    if (
        np.isnan(w_in).any()
        or np.isnan(a_threshold)
        or np.isnan(t_start)
        or np.isnan(walk_forward)
    ):
        return

    if np.floor(t_start) != t_start:
        raise DSPFatal("The starting index must be an integer")

    if np.floor(walk_forward) != walk_forward:
        raise DSPFatal("The search direction must be an integer")

    if int(t_start) < 0 or int(t_start) >= len(w_in):
        raise DSPFatal("The starting index is out of range")

    if int(walk_forward) == 1:
        for i in range(int(t_start), len(w_in) - 1, 1):
            if w_in[i] <= a_threshold < w_in[i + 1]:
                t_out[0] = i
                return
    else:
        for i in range(int(t_start), 0, -1):
            if w_in[i - 1] < a_threshold <= w_in[i]:
                t_out[0] = i
                return


@guvectorize(
    [
        "void(float32[:], float32, float32, int64, char, float32[:])",
        "void(float64[:], float64, float64, int64, char, float64[:])",
    ],
    "(n),(),(),(),()->()",
    **nb_kwargs,
)
def interpolated_time_point_thresh(
    w_in: np.ndarray,
    a_threshold: float,
    t_start: int,
    walk_forward: int,
    mode_in: np.int8,
    t_out: float,
) -> None:
    """Find the time where the waveform value crosses the threshold

    Search performed walking either forward or backward from the starting
    index. Use interpolation to estimate a time between samples. Interpolation
    mode selected with `mode_in`.

    Parameters
    ----------
    w_in
        the input waveform.
    a_threshold
        the threshold value.
    t_start
        the starting index.
    walk_forward
        the backward (``0``) or forward (``1``) search direction.
    mode_in
        Character selecting which interpolation method to use. Note this
        must be passed as a ``int8``, e.g. ``ord('i')``. Options:

        * ``i`` -- integer `t_in`; equivalent to
          :func:`~.dsp.processors.fixed_sample_pickoff`
        * ``b`` -- before; closest integer sample before threshold crossing
        * ``a`` -- after; closest integer sample after threshold crossing
        * ``r`` -- round; round to nearest integer sample to threshold crossing
        * ``l`` -- linear interpolation

        The following modes are meant to mirror the options
        dspeed.upsampler

        * ``f`` -- floor; interpolated values are at previous neighbor.
          Equivalent to ``a``
        * ``c`` -- ceiling, interpolated values are at next neighbor.
          Equivalent to ``b``
        * ``n`` -- nearest-neighbor interpolation; threshold crossing is
          half-way between samples
        * ``h`` -- Hermite cubic spline interpolation (*not implemented*)
        * ``s`` -- natural cubic spline interpolation (*not implemented*)
    t_out
        the index where the waveform value crosses the threshold.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "tp_0": {
            "function": "time_point_thresh",
            "module": "dspeed.processors",
            "args": ["wf_atrap", "bl_std", "tp_start", 0, "'l'", "tp_0"],
            "unit": "ns"
        }
    """
    t_out[0] = np.nan

    if (
        np.isnan(w_in).any()
        or np.isnan(a_threshold)
        or np.isnan(t_start)
        or np.isnan(walk_forward)
    ):
        return

    if t_start < 0 or t_start >= len(w_in):
        return

    i_cross = -1
    if walk_forward > 0:
        for i in range(int(t_start), len(w_in) - 1, 1):
            if w_in[i] <= a_threshold < w_in[i + 1]:
                i_cross = i
                break
    else:
        for i in range(int(t_start), 1, -1):
            if w_in[i - 1] < a_threshold <= w_in[i]:
                i_cross = i - 1
                break

    if i_cross == -1:
        return

    if mode_in == ord("i"):  # return index before crossing
        t_out[0] = i_cross
    elif mode_in in (ord("a"), ord("f")):  # return index after crossing
        t_out[0] = i_cross + 1
    elif mode_in in (ord("b"), ord("c")):  # return index before crossing
        t_out[0] = i_cross
    elif mode_in == ord("r"):  # return closest index to crossing
        if abs(a_threshold - w_in[i_cross]) < abs(a_threshold - w_in[i_cross + 1]):
            t_out[0] = i_cross
        else:
            t_out[0] = i_cross + 1
    elif mode_in == ord("n"):  # nearest-neighbor; return half-way between samps
        t_out[0] = i_cross + 0.5
    elif mode_in == ord("l"):  # linear
        t_out[0] = i_cross + (a_threshold - w_in[i_cross]) / (
            w_in[i_cross + 1] - w_in[i_cross]
        )
    else:
        raise DSPFatal("Unrecognized interpolation mode")


@guvectorize(
    [
        "void(float32[:], float32[:], float32, float32, char, float32[:])",
        "void(float64[:], float64[:], float64, float64, char, float64[:])",
    ],
    "(n),(m),(),(),()->(m)",
    **nb_kwargs,
)
def multi_time_point_thresh(
    w_in: np.ndarray,
    a_threshold: np.ndarray,
    t_start: int,
    polarity: int,
    mode_in: np.int8,
    t_out: np.ndarray,
) -> None:
    """Find the time where the waveform value crosses the threshold

    Search performed walking either forward or backward from the starting
    index. Use interpolation to estimate a time between samples. Interpolation
    mode selected with `mode_in`.

    Parameters
    ----------
    w_in
        the input waveform.
    a_threshold
        list of threshold values.
    t_start
        the starting index.
    polarity
        is the average slope of the waveform up (>0) or down (<0) in the
        search region; only sign matters, not value. Raise Exception if 0.
    mode_in
        Character selecting which interpolation method to use. Note this
        must be passed as a ``int8``, e.g. ``ord('i')``. Options:

        * ``i`` -- integer `t_in`; equivalent to
          :func:`~.dsp.processors.fixed_sample_pickoff`
        * ``b`` -- before; closest integer sample before threshold crossing
        * ``a`` -- after; closest integer sample after threshold crossing
        * ``r`` -- round; round to nearest integer sample to threshold crossing
        * ``l`` -- linear interpolation

        The following modes are meant to mirror the options
        dspeed.upsampler

        * ``f`` -- floor; interpolated values are at previous neighbor.
          Equivalent to ``a``
        * ``c`` -- ceiling, interpolated values are at next neighbor.
          Equivalent to ``b``
        * ``n`` -- nearest-neighbor interpolation; threshold crossing is
          half-way between samples
        * ``h`` -- Hermite cubic spline interpolation (*not implemented*)
        * ``s`` -- natural cubic spline interpolation (*not implemented*)
    t_out
        the index where the waveform value crosses the threshold.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "tp_0": {
            "function": "time_point_thresh",
            "module": "dspeed.processors",
            "args": ["wf_atrap", "bl_std", "tp_start", 0, "'l'", "tp_0"],
            "unit": "ns"
        }
    """
    t_out[:] = np.nan

    if np.isnan(w_in).any() or np.isnan(a_threshold).any() or np.isnan(t_start):
        return

    if t_start < 0 or t_start >= len(w_in):
        return

    # make polarity +/- 1
    if polarity > 0:
        polarity = 1
    elif polarity < 0:
        polarity = -1
    else:
        raise DSPFatal("polarity cannot be 0")

    sorted_idx = np.argsort(a_threshold)

    # Get initial values for search
    t_start = int(t_start)
    a_start = w_in[t_start]
    i_start = len(sorted_idx)
    for i in range(len(sorted_idx)):
        if a_threshold[sorted_idx[i]] >= a_start:
            i_start = i
            break

    # Search for timepoints at larger values
    i_tp = i_start
    if i_tp < len(sorted_idx):
        idx = sorted_idx[i_tp]
        for i_wf in range(t_start, len(w_in) - 1 if polarity > 0 else -1, polarity):
            if i_tp >= len(sorted_idx):
                break
            while w_in[i_wf] <= a_threshold[idx] < w_in[i_wf + polarity]:
                if mode_in == ord("i"):  # return index closest to start of search
                    t_out[idx] = i_wf
                elif mode_in in (ord("a"), ord("f")):  # return index after crossing
                    t_out[idx] = i_wf if polarity < 0 else i_wf + 1
                elif mode_in in (ord("b"), ord("c")):  # return index before crossing
                    t_out[idx] = i_wf if polarity > 0 else i_wf - 1
                elif mode_in == ord("r"):  # round; return closest index
                    if (
                        a_threshold[idx] - w_in[i_wf]
                        < w_in[i_wf + polarity] - a_threshold[sorted_idx[i_tp]]
                    ):
                        t_out[idx] = i_wf
                    else:
                        t_out[idx] = i_wf + polarity
                elif mode_in == ord(
                    "n"
                ):  # nearest-neighbor; return half-way between samps
                    t_out[idx] = i_wf + 0.5 * polarity
                elif mode_in == ord("l"):  # linear
                    t_out[idx] = i_wf + (a_threshold[idx] - w_in[i_wf]) / (
                        w_in[i_wf + polarity] - w_in[i_wf]
                    )
                else:
                    raise DSPFatal("Unrecognized interpolation mode")
                i_tp += 1
                if i_tp >= len(sorted_idx):
                    break
                idx = sorted_idx[i_tp]

    # Search for timepoints at smaller values
    i_tp = i_start - 1
    if i_tp >= 0:
        idx = sorted_idx[i_tp]
        for i_wf in range(
            t_start - 1, len(w_in) - 1 if polarity < 0 else -1, -polarity
        ):
            if i_tp < 0:
                break
            while w_in[i_wf] <= a_threshold[idx] < w_in[i_wf + polarity]:
                if mode_in == ord("i"):  # return index closest to start of search
                    t_out[idx] = i_wf
                elif mode_in in (ord("a"), ord("f")):  # return index after crossing
                    t_out[idx] = i_wf if polarity < 0 else i_wf + 1
                elif mode_in in (ord("b"), ord("c")):  # return index before crossing
                    t_out[idx] = i_wf if polarity > 0 else i_wf - 1
                elif mode_in == ord("r"):  # round; return closest index
                    if (
                        a_threshold[idx] - w_in[i_wf]
                        < w_in[i_wf + polarity] - a_threshold[idx]
                    ):
                        t_out[idx] = i_wf
                    else:
                        t_out[idx] = i_wf + polarity
                elif mode_in == ord(
                    "n"
                ):  # nearest-neighbor; return half-way between samps
                    t_out[idx] = i_wf + 0.5 * polarity
                elif mode_in == ord("l"):  # linear
                    t_out[idx] = i_wf + (a_threshold[idx] - w_in[i_wf]) / (
                        w_in[i_wf + polarity] - w_in[i_wf]
                    )
                else:
                    raise DSPFatal("Unrecognized interpolation mode")
                i_tp -= 1
                if i_tp < 0:
                    break
                idx = sorted_idx[i_tp]


@guvectorize(
    [
        "void(float32[:], float32, float32, float32, float32, uint32[:], float32[:], float32[:])",
        "void(float64[:], float64, float64, float64, float64, uint32[:], float64[:], float64[:])",
    ],
    "(n),(),(),(),(),(),(m),(m)",
    **nb_kwargs,
)
def bi_level_zero_crossing_time_points(
    w_in: np.ndarray,
    a_pos_threshold_in: float,
    a_neg_threshold_in: float,
    gate_time_in: int,
    t_start_in: int,
    n_crossings_out: int,
    polarity_out: np.array,
    t_trig_times_out: np.array,
) -> None:
    """
    Find the indices where a waveform value crosses 0 after crossing the threshold and reaching the next threshold within some gate time.
    Works on positive and negative polarity waveforms.
    Useful for finding pileup events with the RC-CR^2 filter.

    Parameters
    ----------
    w_in
        the input waveform.
    a_pos_threshold_in
        the positive threshold value.
    a_neg_threshold_in
        the negative threshold value.
    gate_time_in
        The number of samples that the next threshold crossing has to be within in order to count a 0 crossing
    t_start_in
        the starting index.
    n_crossings_out
        the number of zero-crossings found. Note: if there are more zeros than elements in output arrays, this will continue to increment but the polarity and trigger time will not be added to the output buffers
    polarity_out
        An array holding the polarity of identified pulses. 0 for negative and 1 for positive
    t_trig_times_out
        the indices where the waveform value has crossed the threshold and returned to 0.
        Arrays of fixed length (padded with :any:`numpy.nan`) that hold the
        indices of the identified trigger times.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "trig_times_out": {
            "function": "multi_trigger_time",
            "module": "dspeed.processors",
            "args": ["wf_rc_cr2", "5", "-10", 0, "n_crossings", "polarity_out(20, vector_len=n_crossings)", "trig_times_out(20, vector_len=n_crossings)"],
            "unit": "ns"
        }
    """
    # prepare output
    t_trig_times_out[:] = np.nan
    polarity_out[:] = np.nan

    # Check everything is ok
    if (
        np.isnan(w_in).any()
        or np.isnan(a_pos_threshold_in)
        or np.isnan(a_neg_threshold_in)
        or np.isnan(t_start_in)
    ):
        return

    if np.floor(t_start_in) != t_start_in:
        raise DSPFatal("The starting index must be an integer")

    if int(t_start_in) < 0 or int(t_start_in) >= len(w_in):
        raise DSPFatal("The starting index is out of range")

    if len(polarity_out) != len(t_trig_times_out):
        raise DSPFatal("The output arrays are of different lengths.")

    gate_time_in = int(gate_time_in)  # make sure this is an integer!
    # Perform the processing
    is_above_thresh = False
    is_below_thresh = False
    crossed_zero = False
    n_crossings_out[0] = 0
    for i in range(int(t_start_in), len(w_in) - 1, 1):
        if is_below_thresh and (w_in[i] <= 0 < w_in[i + 1]):
            crossed_zero = True
            neg_trig_time_candidate = i

        # Either we go above threshold
        if w_in[i] <= a_pos_threshold_in < w_in[i + 1]:
            if crossed_zero and is_below_thresh:
                if i - is_below_thresh < gate_time_in:
                    if n_crossings_out[0] < len(polarity_out):
                        t_trig_times_out[n_crossings_out[0]] = neg_trig_time_candidate
                        polarity_out[n_crossings_out[0]] = 0
                    n_crossings_out[0] += 1
                else:
                    is_above_thresh = i

                is_below_thresh = False
                crossed_zero = False
            else:
                is_above_thresh = i

        if is_above_thresh and (w_in[i] >= 0 > w_in[i + 1]):
            crossed_zero = True
            pos_trig_time_candidate = i

        # Or we go below threshold
        if w_in[i] >= a_neg_threshold_in > w_in[i + 1]:
            if crossed_zero and is_above_thresh:
                if i - is_above_thresh < gate_time_in:
                    if n_crossings_out[0] < len(polarity_out):
                        t_trig_times_out[n_crossings_out[0]] = pos_trig_time_candidate
                        polarity_out[n_crossings_out[0]] = 1
                    n_crossings_out[0] += 1
                else:
                    is_below_thresh = i
                is_above_thresh = False
                crossed_zero = False
            else:
                is_below_thresh = i



import numpy as np
import pytest

from dspeed.errors import DSPFatal
from dspeed.processors import get_wf_centroid


def test_get_wf_centroid(compare_numba_vs_python):
    len_wf = 20

    # test for nan if w_in has a nan
    w_in = np.ones(len_wf)
    w_in[4] = np.nan
    assert np.isnan(compare_numba_vs_python(get_wf_centroid, w_in, 1))

    # test for nan if nan is passed to shift
    w_in = np.ones(len_wf)
    with pytest.raises(DSPFatal):
        compare_numba_vs_python(get_wf_centroid, w_in, np.nan)

    # test for nan if shift is negative
    w_in = np.ones(len_wf)
    with pytest.raises(DSPFatal):
        compare_numba_vs_python(get_wf_centroid, w_in, -1)

    # test for nan if shift is too large
    w_in = np.ones(len_wf)
    with pytest.raises(DSPFatal):
        compare_numba_vs_python(get_wf_centroid, w_in, len_wf)



from __future__ import annotations

import numpy as np
from numba import vectorize

from ..errors import DSPFatal
from ..utils import numba_defaults_kwargs as nb_kwargs

# Functions used for converting between different coordinate systems
# in ProcessingChain. Convert timing values knowing the offset
# in each coordinate frame, and the ratio of periods


@vectorize(
    [f"{t}({t}, f8, f8, f8)" for t in ["f4", "f8"]],
    **nb_kwargs,
)
def convert(buf_in, offset_in, offset_out, period_ratio):  # noqa: N805
    return (buf_in + offset_in) * period_ratio - offset_out


@vectorize(
    [f"{t}({t}, f8, f8, f8)" for t in ["u1", "u2", "u4", "u8", "i1", "i2", "i4", "i8"]],
    **nb_kwargs,
)
def convert_int(buf_in, offset_in, offset_out, period_ratio):  # noqa: N805
    tmp = (buf_in + offset_in) * period_ratio - offset_out
    ret = np.rint(tmp)
    if np.abs(tmp - ret) < 1.0e-5:
        return ret
    else:
        raise DSPFatal("Cannot convert to integer. Use round or astype")


@vectorize(
    [
        f"{t}({t}, f8, f8, f8)"
        for t in ["u1", "u2", "u4", "u8", "i1", "i2", "i4", "i8", "f4", "f8"]
    ],
    **nb_kwargs,
)
def convert_round(buf_in, offset_in, offset_out, period_ratio):  # noqa: N805
    return np.rint((buf_in + offset_in) * period_ratio - offset_out)


@vectorize(
    [
        f"{t}({t}, f8, f8, f8)"
        for t in ["u1", "u2", "u4", "u8", "i1", "i2", "i4", "i8", "f4", "f8"]
    ],
    **nb_kwargs,
)
def convert_floor(buf_in, offset_in, offset_out, period_ratio):  # noqa: N805
    return np.floor((buf_in + offset_in) * period_ratio - offset_out)


@vectorize(
    [
        f"{t}({t}, f8, f8, f8)"
        for t in ["u1", "u2", "u4", "u8", "i1", "i2", "i4", "i8", "f4", "f8"]
    ],
    **nb_kwargs,
)
def convert_ceil(buf_in, offset_in, offset_out, period_ratio):  # noqa: N805
    return np.ceil((buf_in + offset_in) * period_ratio - offset_out)


@vectorize(
    [
        f"{t}({t}, f8, f8, f8)"
        for t in ["u1", "u2", "u4", "u8", "i1", "i2", "i4", "i8", "f4", "f8"]
    ],
    **nb_kwargs,
)
def convert_trunc(buf_in, offset_in, offset_out, period_ratio):  # noqa: N805
    return np.trunc((buf_in + offset_in) * period_ratio - offset_out)



import numpy as np
import pytest

from dspeed.errors import DSPFatal
from dspeed.processors import (
    bi_level_zero_crossing_time_points,
    interpolated_time_point_thresh,
    rc_cr2,
    time_point_thresh,
)


def test_time_point_thresh(compare_numba_vs_python):
    """Testing function for the time_point_thresh processor."""

    # test for nan if w_in has a nan
    w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
    w_in[4] = np.nan
    assert np.isnan(
        compare_numba_vs_python(
            time_point_thresh,
            w_in,
            1,
            11,
            0,
        )
    )

    # test for nan if nan is passed to a_threshold
    w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
    assert np.isnan(
        compare_numba_vs_python(
            time_point_thresh,
            w_in,
            np.nan,
            11,
            0,
        )
    )

    # test for nan if nan is passed to t_start
    w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
    assert np.isnan(
        compare_numba_vs_python(
            time_point_thresh,
            w_in,
            1,
            np.nan,
            0,
        )
    )

    # test for nan if nan is passed to walk_forward
    w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
    assert np.isnan(
        compare_numba_vs_python(
            time_point_thresh,
            w_in,
            1,
            11,
            np.nan,
        )
    )

    # test for error if t_start non integer
    with pytest.raises(DSPFatal):
        w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
        time_point_thresh(w_in, 1, 10.5, 0, np.array([0.0]))

    # test for error if walk_forward non integer
    with pytest.raises(DSPFatal):
        w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
        time_point_thresh(w_in, 1, 11, 0.5, np.array([0.0]))

    # test for error if t_start out of range
    with pytest.raises(DSPFatal):
        w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
        time_point_thresh(w_in, 1, 12, 0, np.array([0.0]))

    # test walk backward
    w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
    assert compare_numba_vs_python(time_point_thresh, w_in, 1, 11, 0) == 8.0

    # test walk forward
    w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
    assert compare_numba_vs_python(time_point_thresh, w_in, 3, 0, 1) == 4.0


def test_interpolated_time_point_thresh(compare_numba_vs_python):
    """Testing function for the interpolated_time_point_thresh processor."""

    # test for nan if w_in has a nan
    w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
    w_in[4] = np.nan
    assert np.isnan(
        compare_numba_vs_python(interpolated_time_point_thresh, w_in, 1.0, 11.0, 0, 105)
    )

    # test for nan if nan is passed to a_threshold
    w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
    assert np.isnan(
        compare_numba_vs_python(
            interpolated_time_point_thresh, w_in, np.nan, 11.0, 0, 105
        )
    )

    # test for nan if nan is passed to t_start
    w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
    assert np.isnan(
        compare_numba_vs_python(
            interpolated_time_point_thresh, w_in, 1.0, np.nan, 0, 105
        )
    )

    # test for nan if t_start out of range
    w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
    assert np.isnan(
        compare_numba_vs_python(interpolated_time_point_thresh, w_in, 1.0, 12, 0, 105)
    )

    # test walk backward mode 'i'
    w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
    assert (
        compare_numba_vs_python(interpolated_time_point_thresh, w_in, 1, 11, 0, 105)
        == 7.0
    )

    # test walk forward mode 'i'
    w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
    assert (
        compare_numba_vs_python(interpolated_time_point_thresh, w_in, 3, 0, 1, 105)
        == 4.0
    )

    # test walk backward mode 'f'
    w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
    assert (
        compare_numba_vs_python(interpolated_time_point_thresh, w_in, 1, 11, 0, 102)
        == 8.0
    )

    # test walk forward mode 'f'
    w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
    assert (
        compare_numba_vs_python(interpolated_time_point_thresh, w_in, 3, 0, 1, 102)
        == 5.0
    )

    # test walk backward mode 'f'
    w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
    assert (
        compare_numba_vs_python(interpolated_time_point_thresh, w_in, 1, 11, 0, 99)
        == 7.0
    )

    # test walk forward mode 'f'
    w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
    assert (
        compare_numba_vs_python(interpolated_time_point_thresh, w_in, 3, 0, 1, 99)
        == 4.0
    )

    # test walk backward mode 'n'
    w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
    assert (
        compare_numba_vs_python(interpolated_time_point_thresh, w_in, 1, 11, 0, 110)
        == 7.5
    )

    # test walk forward mode 'n'
    w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
    assert (
        compare_numba_vs_python(interpolated_time_point_thresh, w_in, 3, 0, 1, 110)
        == 4.5
    )

    # test walk backward mode 'l'
    w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
    assert (
        compare_numba_vs_python(interpolated_time_point_thresh, w_in, 1.5, 11, 0, 108)
        == 8.5
    )

    # test walk forward mode 'l'
    w_in = np.concatenate([np.arange(-1, 5, 1), np.arange(-1, 5, 1)], dtype="float")
    assert (
        compare_numba_vs_python(interpolated_time_point_thresh, w_in, 3.5, 0, 1, 108)
        == 4.5
    )


def test_bi_level_zero_crossing_time_points(compare_numba_vs_python):
    # Test exceptions and initial checks
    # ensure that if there is a nan in w_in, all nans are outputted
    w_in = np.ones(100)
    w_in[4] = np.nan
    n_out = np.zeros(1, "uint32")
    t_out = np.zeros(5)
    pol_out = np.zeros(5)
    bi_level_zero_crossing_time_points(w_in, 100, 100, 100, 0, n_out, pol_out, t_out)
    assert np.isnan(t_out).all()

    # ensure the ValueError is raised if the polarity output array is different length than the time point output array
    t_start_in = 1.02
    with pytest.raises(ValueError):
        bi_level_zero_crossing_time_points(
            np.ones(9), 100, 100, 100, t_start_in, n_out, pol_out, np.zeros(1)
        )

    # ensure the DSPFatal is raised if initial timepoint is not an integer
    t_start_in = 1.02
    with pytest.raises(DSPFatal):
        bi_level_zero_crossing_time_points(
            np.ones(9), 100, 100, 100, t_start_in, n_out, pol_out, t_out
        )

    # ensure the DSPFatal is raised if initial timepoint is not negative
    t_start_in = -2
    with pytest.raises(DSPFatal):
        bi_level_zero_crossing_time_points(
            np.ones(9), 100, 100, 100, t_start_in, n_out, pol_out, t_out
        )

    # ensure the DSPFatal is raised if initial timepoint is outside length of waveform
    t_start_in = 100
    with pytest.raises(DSPFatal):
        bi_level_zero_crossing_time_points(
            np.ones(9), 100, 100, 100, t_start_in, n_out, pol_out, t_out
        )

    early_trig = 500  # start pulse1 500 samples from the start of the wf
    late_trig = 200  # start pulse2 200 samples after the midpoint of 8192 length wf
    zeta = 30000  # the decay time constant of a pulse, in samples
    amplitude = 1750
    tau = 100  # the RC-CR^2 filter time constant

    # Make the first pulse and RC-CR^2 filter it
    ts = np.arange(0, 8192 // 2 - early_trig)
    pulse = amplitude * np.exp(-1 * ts / zeta)
    pulse = np.insert(pulse, 0, np.zeros(5))  # pad with 0s to avoid edge effects
    out_pulse = np.zeros_like(pulse)

    rc_cr2(pulse, tau, out_pulse)

    pulse = np.insert(
        out_pulse[5:], 0, np.zeros(early_trig)
    )  # delay the start of the wf by early_trig

    # Make the second pulse
    ts = np.arange(0, 8192 // 2 - late_trig)
    pulse2 = amplitude * np.exp(-1 * ts / zeta)
    pulse2 = np.insert(pulse2, 0, np.zeros(5))  # avoid edge effects
    out_pulse = np.zeros_like(pulse2)
    rc_cr2(pulse2, tau, out_pulse)
    pulse = np.insert(pulse, -1, np.zeros(late_trig))
    pulse = np.insert(pulse, -1, out_pulse[5:])

    gate_time = 1000
    # Test that the filter reproduces 0 crossings at the expected points
    # Test on positive polarity
    t_trig_times_out = np.zeros(5)
    pol_out = np.zeros(5)
    bi_level_zero_crossing_time_points(
        pulse, 2000, -2000, gate_time, 0, n_out, pol_out, t_trig_times_out
    )

    cross_1 = early_trig + 2 * tau - 1  # minus 1 from delay?
    cross_2 = (
        8192 // 2 + late_trig + 2 * tau - 2
    )  # minus 1 from 1st pulse delay and again
    assert np.allclose(int(t_trig_times_out[0]), cross_1, rtol=1)
    assert np.allclose(int(t_trig_times_out[1]), cross_2, rtol=1)
    assert int(pol_out[0]) == 1
    assert int(pol_out[1]) == 1

    # Check on negative polarity pulses
    t_trig_times_out = np.zeros(5)
    pol_out = np.zeros(5)
    bi_level_zero_crossing_time_points(
        -1 * pulse, 2000, -2000, gate_time, 0, n_out, pol_out, t_trig_times_out
    )
    assert np.allclose(int(t_trig_times_out[0]), cross_1, rtol=1)
    assert np.allclose(int(t_trig_times_out[1]), cross_2, rtol=1)
    assert int(pol_out[0]) == 0
    assert int(pol_out[1]) == 0

    # Check positive polarity pulses that cross 0 and never reach negative threshold return all nan
    t_trig_times_out = np.zeros(5)
    pol_out = np.zeros(5)
    bi_level_zero_crossing_time_points(
        pulse, 2000, -300000, gate_time, 0, n_out, pol_out, t_trig_times_out
    )
    assert np.isnan(t_trig_times_out).all()
    assert np.isnan(pol_out).all()

    # Check negative polarity pulses that cross 0 and never reach positive threshold return all nan
    t_trig_times_out = np.zeros(5)
    pol_out = np.zeros(5)
    bi_level_zero_crossing_time_points(
        -1 * pulse, 300000, -2000, gate_time, 0, n_out, pol_out, t_trig_times_out
    )
    assert np.isnan(t_trig_times_out).all()
    assert np.isnan(pol_out).all()

    # Check that pulses that never reach either threshold return all nan
    t_trig_times_out = np.zeros(5)
    pol_out = np.zeros(5)
    bi_level_zero_crossing_time_points(
        pulse, 300000, 300000, gate_time, 0, n_out, pol_out, t_trig_times_out
    )
    assert np.isnan(t_trig_times_out).all()
    assert np.isnan(pol_out).all()

    # Check that pulses that go up and never cross zero again return all nan
    t_trig_times_out = np.zeros(5)
    pol_out = np.zeros(5)
    bi_level_zero_crossing_time_points(
        np.linspace(-1, 100, 101), 4, -4, gate_time, 0, n_out, pol_out, t_trig_times_out
    )
    assert np.isnan(t_trig_times_out).all()
    assert np.isnan(pol_out).all()

    # Check that pulses that go down and never cross zero again return all nan
    t_trig_times_out = np.zeros(5)
    pol_out = np.zeros(5)
    bi_level_zero_crossing_time_points(
        -1 * np.linspace(-1, 100, 101),
        4,
        -4,
        gate_time,
        0,
        n_out,
        pol_out,
        t_trig_times_out,
    )
    assert np.isnan(t_trig_times_out).all()
    assert np.isnan(pol_out).all()

    # Check positive polarity pulses where only 2nd peak crosses the threshold
    scale_arr = np.full(8192 // 2, 1)
    scale_arr = np.insert(scale_arr, -1, np.full(8192 // 2, 5))
    t_trig_times_out = np.zeros(5)
    pol_out = np.zeros(5)
    bi_level_zero_crossing_time_points(
        pulse * scale_arr, 2000, -20000, gate_time, 0, n_out, pol_out, t_trig_times_out
    )
    assert np.allclose(
        int(t_trig_times_out[0]), cross_2, rtol=1
    )  # only the 2nd time point should have been crossed
    assert int(pol_out[0]) == 1

    # Check positive polarity pulses where only 1st peak crosses the threshold
    scale_arr = np.full(8192 // 2, 5)
    scale_arr = np.insert(scale_arr, -1, np.full(8192 // 2, 1))
    t_trig_times_out = np.zeros(5)
    pol_out = np.zeros(5)
    bi_level_zero_crossing_time_points(
        pulse * scale_arr, 2000, -20000, gate_time, 0, n_out, pol_out, t_trig_times_out
    )
    assert np.allclose(
        int(t_trig_times_out[0]), cross_1, rtol=1
    )  # only the 1st time point should have been crossed
    assert int(pol_out[0]) == 1

    # Check positive polarity pulses where only 2nd peak crosses both thresholds, but 1st peak passes negative but not within gate
    scale_arr = np.full(8192 // 2, 1)
    scale_arr = np.insert(scale_arr, -1, np.full(8192 // 2, 5))
    t_trig_times_out = np.zeros(5)
    pol_out = np.zeros(5)
    bi_level_zero_crossing_time_points(
        pulse * scale_arr, 50000, -2000, gate_time, 0, n_out, pol_out, t_trig_times_out
    )
    assert np.allclose(
        int(t_trig_times_out[0]), cross_2, rtol=1
    )  # only the 2nd time point should have been crossed
    assert int(pol_out[0]) == 1

    # Check for overflow. This should not fail and should return n_crossings greater than the length of the outputs
    w_in = np.zeros(100, "float") * 10
    w_in[::2] = -10
    w_in[1::2] = 10
    pol_out = np.zeros(30)
    t_trig_out = np.zeros(30)
    bi_level_zero_crossing_time_points(w_in, 1, 1, 10, 0, n_out, pol_out, t_trig_out)
    assert np.all(pol_out == 0)
    assert n_out[0] == 49



import numpy as np

from dspeed.processors import min_max_norm


def test_min_max_norm(compare_numba_vs_python):
    """Testing function for the max_min_norm processor."""

    # set up values to use for each test case
    len_wf = 10

    # ensure that if there is a nan in w_in, all nans are outputted
    w_in = np.ones(len_wf)
    w_in[4] = np.nan
    a = np.array([1])
    assert np.all(np.isnan(compare_numba_vs_python(min_max_norm, w_in, a, a)))

    # test for division by 0
    w_in = np.ones(len_wf)
    a = np.array([0])
    w_out_expected = np.ones(len_wf)
    assert np.allclose(
        compare_numba_vs_python(min_max_norm, w_in, a, a), w_out_expected
    )

    # test for abs(a_max) > abs(a_min)
    a_max = np.array([2])
    a_min = np.array([-1])
    w_out_expected = np.ones(len_wf) / abs(a_max[0])
    assert np.allclose(
        compare_numba_vs_python(min_max_norm, w_in, a_min, a_max), w_out_expected
    )

    # test for abs(a_max) < abs(a_min)
    a_max = np.array([1])
    a_min = np.array([-2])
    w_out_expected = np.ones(len_wf) / abs(a_min[0])
    assert np.allclose(
        compare_numba_vs_python(min_max_norm, w_in, a_min, a_max), w_out_expected
    )



from __future__ import annotations

import numpy as np
from numba import guvectorize

from ..errors import DSPFatal
from ..utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    [
        "void(float32[::1], float32[:], float32[:], float32[::1])",
        "void(float64[::1], float64[:], float64[:], float64[::1])",
    ],
    "(n),(m),(l)->(n)",
    **nb_kwargs,
)
def transfer_function_convolver(w_in, a_in, b_in, w_out):
    r"""
    Compute the difference equation of the form
    $a_0*w_out[i] = a_1*w_out[i-1] + \ldots b_0*w_in[i] \ldots$
    which is equivalent of convolving a signal with a transfer function.
    a_in always needs at least one element
    The numba signature specifies these arrays as contiguous, so that way the dot product is as fast as possible

    Parameters
    ----------
    w_in
        An array of the values that this lti filter will be performed on
    a_in
        An array [a_0, ..., a_n] of filter coefficients to apply recursively to the output
    b_in
        An array [b_0, ..., b_m] of filter coefficients to apply to the input
    w_out
        The output array to store all the filtered values

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "wf_pz": {
            "function": "transfer_function_convolver",
            "module": "dspeed.processors",
            "args": ["waveform", "np.array([-1,1])", "np.array([1,2])", "wf_pz"],
            "unit": "ADC"
        }
    """
    w_out[:] = np.nan

    # perform some checks
    if np.isnan(w_in).any() or np.isnan(a_in).any() or np.isnan(b_in).any():
        return
    if len(w_in) != len(w_out):
        return
    if len(a_in) < 1:
        raise DSPFatal("The parameter a_in must be an array of length at least 1")

    b = b_in / a_in[0]
    a = a_in / a_in[0]

    # copy and flip the arrays, needed so that way they are contiguous
    a_flip = np.copy(a[::-1])
    b_flip = np.copy(b[::-1])

    # Figure out if it's a FIR
    if a_flip.size == 1:
        # Perform the convolution
        for n in range(0, w_in.size):
            w_out[n] = np.dot(b_flip[-(n + 1) :], w_in[n + 1 - len(b_flip) : n + 1])

    # Otherwise we have an IIR
    else:
        a_flip = a_flip[:-1]
        # Initialize the output with the first input signals
        for i in range(a_flip.size):
            w_out[i] = w_in[i]

        # Perform the convolution
        for n in range(a_flip.size, w_in.size):
            w_out[n] = np.dot(a_flip, w_out[n - len(a_flip) : n])
            w_out[n] += np.dot(b_flip[-(n + 1) :], w_in[n + 1 - len(b_flip) : n + 1])



import numpy as np
import pytest

from dspeed.errors import DSPFatal
from dspeed.processors import multi_a_filter


def test_multi_a_filter(compare_numba_vs_python):
    """Testing function for the multi_a_filter."""

    len_wf = 20

    # test for nan if w_in has a nan
    w_in = np.ones(len_wf)
    w_in[4] = np.nan
    assert np.isnan(
        compare_numba_vs_python(multi_a_filter, w_in, np.array([1, 2]))
    ).all()

    w_in = np.ones(len_wf)

    # test vt_maxs_in > wf len
    with pytest.raises(DSPFatal):
        maxs_in = np.arange(21)
        multi_a_filter(w_in, maxs_in, np.zeros(21))

    # test vt_max_len is 0
    assert len(compare_numba_vs_python(multi_a_filter, w_in, np.array([]))) == 0

    # test all nan
    assert np.isnan(
        compare_numba_vs_python(multi_a_filter, w_in, np.array([np.nan]))
    ).all()

    # test output
    assert compare_numba_vs_python(multi_a_filter, w_in, np.array([10]))[0] == 1

    # test output with remaining nans
    out = compare_numba_vs_python(multi_a_filter, w_in, np.array([10, np.nan, np.nan]))
    assert out[0] == 1
    assert np.isnan(out[1:]).all()

    # test output with 1 nan
    out = compare_numba_vs_python(multi_a_filter, w_in, np.array([10, np.nan, 12]))
    assert out[0] == 1 and out[2] == 1
    assert np.isnan(out[1])



from __future__ import annotations

import os
from abc import ABCMeta
from collections.abc import Callable, Collection, Iterator, MutableMapping
from typing import Any

import numpy as np
from numba.np.ufunc import sigparse


class GUFuncWrapper:
    """
    A wrapper class to create a u-func like object from an arbitrary function.
    This class is callable and is intended for use for processors that require
    setup with persistent state information; these processors are generated
    using the "factory" method and typically utilize "init_args"

    Example 1:
    ----------
    .. highlight::python
    .. code-block:: python

        # set up some object 'obj' that has a function we want to call on w_in
        gufunc = GUFuncWrapper(
            lambda w_in: obj.execute(w_in, args...),
            "(n)->()",
            "ff"
        )

    Example 2:
    ----------
    .. highlight::python
    .. code-block:: python

        # fun is a vectorized python function, but we want to use ufunc interface
        gufunc = GUFuncWrapper(
            lambda w_in, a, w_out: fun(w_in, a, out=w_out, ...more kwargs),
            "(n),()->(n)",
            "fff",
            vectorized=True,
            copy_out=False
        )

    """

    def __init__(
        self,
        fun: Callable,
        signature: str,
        types: str | Collection[str],
        name: str | None = None,
        vectorized: bool = False,
        copy_out: bool = True,
        doc_string: str = None,
    ):
        """
        Parameters
        ----------
        fun
            python function to be wrapped
        signature
            gufunction signature (see https://numpy.org/doc/2.1/reference/c-api/generalized-ufuncs.html)
        types
            string of type chars, e.g. fi->f
        name
            name of function. By default use `fun.__name__` (this can be very
            unhelpful, e.g. "<lambda>")
        vectorized
            if False, use np.vectorize to loop over function. Set to True
            if `fun` is already vectorized
        copy_out
            set to False if function does in-place calculation for outputs.
            Cannot be False if vectorized is also False
        doc_string
            manually set doc string. If None, use docstring of fun if it
            exists. Else use this docstring.
        """
        self.__name__ = name if name else fun.__name__
        self.ufunc = fun
        self.signature = signature
        if "->" in signature:
            # numba signature parser can't handle outputs with new dimensions
            self.in_dims, self.out_dims = (
                np.lib._function_base_impl._parse_gufunc_signature(signature)
            )
        else:
            # numpy signature parser can't handle no outputs
            self.in_dims, self.out_dims = sigparse.parse_signature(signature)

        self.nin = len(self.in_dims)
        self.nout = len(self.out_dims)
        self.nargs = self.nin + self.nout
        self.types = [types] if isinstance(types, str) else types
        self.ntypes = 1
        self.copy_out = copy_out
        self.vectorized = vectorized
        if doc_string:
            self.__doc__ = doc_string
        elif fun.__doc__:
            self.__doc__ = fun.__doc__

    def __call__(self, *args):
        """Call wrapped function with "in place" outputs"""

        assert len(args) == self.nargs

        if self.vectorized and self.copy_out and self.nout > 0:
            ins = args[: self.nin]
            outs = args[-self.nout :]
            rets = self.ufunc(*ins)
            if self.nout == 1:
                rets = [rets]
            for out, ret in zip(outs, rets):
                out[...] = ret
        elif self.vectorized:
            self.ufunc(*args)
        else:
            # Calculate broadcasted shapes based on signature
            args = [
                arg if isinstance(arg, np.ndarray) else np.array(arg) for arg in args
            ]

            broadcast_shape, dim_sizes = (
                np.lib._function_base_impl._parse_input_dimensions(
                    args, self.in_dims + self.out_dims
                )
            )
            # handle special case, if scalars were passed as length 1 arrays
            if len(broadcast_shape) > 0 and broadcast_shape[0] == 1:
                broadcast_shape = broadcast_shape[1:]
            shapes = np.lib._function_base_impl._calculate_shapes(
                broadcast_shape, dim_sizes, self.in_dims + self.out_dims
            )

            # check that dimensions match; broadcast inputs that don't
            for i, (arg, shape) in enumerate(zip(args, shapes)):
                if arg.shape != shape and arg.shape != (1,) + shape:
                    if i >= self.nin:
                        raise ValueError("Outputs are not the right shape")
                    args[i] = np.broadcast_to(arg, shape, subok=True)

            # loop over outer dimensions and call function
            for index in np.ndindex(*broadcast_shape):
                ins = [arg[index] for arg in args[: self.nin]]
                outs = []
                for arg in args[self.nin :]:
                    if len(arg.shape) == len(index):
                        if len(index) == 1:
                            outs += [arg[index[0] : index[0] + 1]]
                        else:
                            outs += [arg[np.s_[index[:-1], index[-1] : index[-1] + 1]]]
                    else:
                        outs += [arg[index]]

                if self.copy_out and self.nout > 0:
                    rets = self.ufunc(*ins)
                    if self.nout == 1:
                        rets = [rets]
                    for out, ret in zip(outs, rets):
                        out[...] = ret
                else:
                    self.ufunc(*ins, *outs)


def dspeed_guvectorize(*args, **kwargs):
    """
    Decorator to create a callable object implementing the gufunc interface.
    See arguments in GUFuncWrapper initializer
    """
    return lambda fun: GUFuncWrapper(fun, *args, **kwargs)


def getenv_bool(name: str, default: bool = False) -> bool:
    """Get environment value as a boolean, returning True for 1, t and true
    (caps-insensitive), and False for any other value and default if undefined.
    """
    val = os.getenv(name)
    if not val:
        return default
    elif val.lower() in ("1", "t", "true"):
        return True
    else:
        return False


class NumbaDefaults(MutableMapping):
    """Bare-bones class to store some Numba default options. Defaults values
    are set from environment variables

    Examples
    --------
    Set all default option values for a processor at once by expanding the
    provided dictionary:

    >>> from numba import guvectorize
    >>> from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs
    >>> @guvectorize([], "", **nb_kwargs, nopython=True) # def proc(...): ...

    Customize one argument but still set defaults for the others:

    >>> from pygama.dsp.utils import numba_defaults as nb_defaults
    >>> @guvectorize([], "", **nb_defaults(cache=False) # def proc(...): ...

    Override global options at runtime:

    >>> from pygama.dsp.utils import numba_defaults
    >>> from pygama.dsp import build_dsp
    >>> # must set options before explicitly importing pygama.dsp.processors!
    >>> numba_defaults.cache = False
    >>> numba_defaults.boundscheck = True
    >>> build_dsp(...) # if not explicit, processors imports happen here
    """

    def __init__(self) -> None:
        self.cache: bool = getenv_bool("DSPEED_CACHE", default=True)
        self.boundscheck: bool = getenv_bool("DSPEED_BOUNDSCHECK")
        self.target: str = os.getenv("DSPEED_TARGET", default="cpu")

    def __getitem__(self, item: str) -> Any:
        return self.__dict__[item]

    def __setitem__(self, item: str, val: Any) -> None:
        self.__dict__[item] = val

    def __delitem__(self, item: str) -> None:
        del self.__dict__[item]

    def __iter__(self) -> Iterator:
        return self.__dict__.__iter__()

    def __len__(self) -> int:
        return len(self.__dict__)

    def __call__(self, **kwargs) -> dict:
        mapping = self.__dict__.copy()
        mapping.update(**kwargs)
        return mapping

    def __str__(self) -> str:
        return str(self.__dict__)

    def __repr__(self) -> str:
        return str(self.__dict__)


numba_defaults = NumbaDefaults()
numba_defaults_kwargs = numba_defaults


class ProcChainVarBase(metaclass=ABCMeta):
    r"""Base class.

    :class:`ProcChainVar` implements this class. This base class is used
    by processors that use ProcChainVar in their constructors.
    """

    pass



from __future__ import annotations

import numpy as np
from numba import guvectorize

from dspeed.errors import DSPFatal
from dspeed.utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    [
        "void(float32[:], float32, float32[:])",
        "void(float64[:], float64, float64[:])",
    ],
    "(n),()->()",
    **nb_kwargs,
)
def get_wf_centroid(w_in: np.ndarray, shift: int, centroid: int) -> None:
    """Calculate waveform centroid.

    Note
    ----
    This processor calculate the centroid position when provided the convolution product with a step function.

    Parameters
    ----------
    w_in
        the input waveform.
    shift
        shift.
    centroid
        centroid position.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "centroid": {
          "function": "get_wf_centroid",
          "module": "dspeed.processors",
          "args": ["waveform", "shift", "centroid"],
          "unit": "ADC"
        }
    """

    centroid[0] = np.nan

    if np.isnan(w_in).any():
        return

    if np.isnan(shift):
        raise DSPFatal("shift is nan")
    if shift < 0:
        raise DSPFatal("shift must be positive")
    if shift > len(w_in) - 1:
        raise DSPFatal("shift must be shorter than input waveform size")

    c_a = (
        np.where(w_in[w_in.argmin() : w_in.argmax()] > 0)[0][0] + w_in.argmin() + shift
    )
    c_b = (
        np.where(w_in[w_in.argmin() : w_in.argmax()] < 0)[0][-1] + w_in.argmin() + shift
    )

    centroid[0] = round((c_a + c_b) / 2)



from __future__ import annotations

import numpy as np
from numba import guvectorize

from ..errors import DSPFatal
from ..utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    [
        "void(float32[:,::1], float32[::1], float32, float32, boolean, float32, float32[::1])",
        "void(float64[:,::1], float64[::1], float64, float32, boolean, float32, float64[::1])",
    ],
    "(m,n),(m),(),(),(),(),(n)",
    nopython=True,
    **nb_kwargs,
)
def optimize_nnls(
    a: np.ndarray,
    b: np.ndarray,
    maxiter: int,
    tol: float,
    allow_singularity: bool,
    min_value: float,
    x: np.ndarray,
) -> None:
    """Solve ``argmin_x || ax - b ||_2`` for ``x>=0``.
    Based on :func:`scipy.optimize.nnls` implementation. Which in turn is based on
    the algorithm in Bro, R. and De Jong, S. (1997), A fast non-negativity-constrained least squares algorithm. J. Chemometrics, 11: 393-401

    Parameters
    ----------
    a : (m, n) ndarray
        Coefficient matrix
    b : (m,) ndarray, float
        Right-hand side vector.
    maxiter: int
        Maximum number of iterations.
    tol: float
        Tolerance value used in the algorithm to assess closeness to zero in
        the projected residual ``(a.T @ (a x - b)`` entries. Increasing this
        value relaxes the solution constraints.
    allow_singularity: bool
        If matrix is not solvable (e.g. because of non full rank caused by
        float precision), no error is raised but all elements of the
        solution vector are set NaN
    x : ndarray
        Solution vector.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "nnls_solution": {
            "function": "optimize_nnls",
            "module": "dspeed.processors",
            "args": ["db.coefficient_matrix",
                "wf_blsub",
                "1000", "1e-6", "True"
                "nnls_solution"],
        }
    """

    def numba_ix(arr: np.array, rows: np.array, cols: np.array) -> np.array:
        """
        Numba compatible implementation of arr[np.ix_(rows, cols)] for 2D arrays.
        from https://github.com/numba/numba/issues/5894#issuecomment-974701551
        :param arr: 2D array to be indexed
        :param rows: Row indices
        :param cols: Column indices
        :return: 2D array with the given rows and columns of the input array
        """
        one_d_index = np.zeros(len(rows) * len(cols), dtype=np.int32)
        for i, r in enumerate(rows):
            start = i * len(cols)
            one_d_index[start : start + len(cols)] = cols + arr.shape[1] * r

        arr_1d = arr.reshape((arr.shape[0] * arr.shape[1], 1))
        slice_1d = np.take(arr_1d, one_d_index)
        return slice_1d.reshape((len(rows), len(cols)))

    def is_singular(matrix):
        """
        Returns True if matrix det = 0 i.e. matrix is singular.
        """
        det = np.linalg.det(matrix)
        return abs(det) < np.finfo(np.float64).eps

    m, n = a.shape

    if n != len(x):
        raise DSPFatal(
            "n dimension of coefficient axis doesn't match solution vector length."
        )
    if m != len(b):
        raise DSPFatal(
            "m dimension of coefficient axis doesn't match right-hand vector length."
        )

    ata = np.transpose(a) @ a
    atb = b @ a  # Result is 1D - let NumPy figure it out

    # Initialize vars
    x[:] = np.zeros(n, dtype=np.float64)
    s = np.zeros(n, dtype=np.float64)
    # Inactive constraint switches
    p = np.zeros(n, dtype=np.bool_)
    pidx = np.arange(0, len(p), 1, dtype=np.int32)

    # Projected residual
    w = atb.copy().astype(np.float64)  # x=0. Skip (-ata @ x) term

    # Overall iteration counter
    # Outer loop is not counted, inner iter is counted across outer spins
    iter = 0
    while (not p.all()) and (w[~p] > tol).any():
        # Get the "most" active coeff index and move to inactive set
        k = np.argmax(w * (~p))
        p[k] = True

        # Iteration solution
        s[:] = 0.0
        mat = numba_ix(ata, pidx[p], pidx[p])

        # check if matrix has full rank before solving
        if is_singular(mat) and allow_singularity:
            x[:] = np.nan
            return None

        s[p] = np.linalg.solve(mat, atb[p])

        # Inner loop
        while (iter < maxiter) and (s[p].min() <= min_value):
            iter += 1
            inds = p * (s <= min_value)
            alpha = (x[inds] / (x[inds] - s[inds])).min()
            x *= 1 - alpha
            x += alpha * s
            p[x <= tol] = False

            mat = numba_ix(ata, pidx[p], pidx[p])
            if is_singular(mat) and allow_singularity:
                x[:] = np.nan
                return None

            s[p] = np.linalg.solve(mat, atb[p])
            s[~p] = 0

        x[:] = s[:]
        w[:] = atb - ata @ x

        if iter == maxiter:
            return None



from __future__ import annotations

import numpy as np
from numba import guvectorize

from ..errors import DSPFatal
from ..utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    ["void(float32[:], float32, float32[:])", "void(float64[:], float64, float64[:])"],
    "(n),()->(n)",
    **nb_kwargs,
)
def rc_cr2(w_in: np.array, t_tau: float, w_out: np.array) -> None:
    """
    Apply a RC-CR^2 filter with the provided time
    constant to the waveform. Useful for determining pileup
    and trigger times. The filter was computed using a matched z-transform
    to keep the poles/zeroes of the analog transfer function in the same location.

    Parameters
    ----------
    w_in
        the input waveform.
    t_tau
        the time constant of the integration and differentiaion
    w_out
        the RC-CR^2 filtered waveform.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "wf_RC_CR2": {
            "function": "rc_cr2",
            "module": "dspeed.processors",
            "args": ["wf_bl", "300*ns", "wf_RC_CR2"],
            "unit": "ADC"
        }
    """
    w_out[:] = np.nan

    if np.isnan(w_in).any() or np.isnan(t_tau):
        return

    if len(w_in) <= 3:
        raise DSPFatal(
            "The length of the waveform must be larger than 3 for the filter to work safely"
        )

    w_out[0] = w_in[0]
    w_out[1] = w_in[1]
    w_out[2] = w_in[2]

    # Make a temporary buffer at higher precision so that the recursive filter doesn't compound any float truncations
    w_tmp = np.zeros(4, dtype=np.float64)
    w_tmp[0] = w_in[0]
    w_tmp[1] = w_in[1]
    w_tmp[2] = w_in[2]

    a = np.exp(-1 / t_tau)

    denom_1 = 1
    denom_2 = -3 * a
    denom_3 = 3 * a**2
    denom_4 = -(a**3)

    num_1 = 1
    num_2 = -2
    num_3 = 1

    for i in range(3, len(w_in)):
        w_tmp[3] = (
            -denom_2 * w_tmp[2]
            - denom_3 * w_tmp[1]
            - denom_4 * w_tmp[0]
            + num_1 * w_in[i]
            + num_2 * w_in[i - 1]
            + num_3 * w_in[i - 2]
        ) / denom_1
        w_out[i] = w_tmp[3]  # Put the higher precision buffer into the desired output
        # shuffle the buffers
        w_tmp[0] = w_tmp[1]
        w_tmp[1] = w_tmp[2]
        w_tmp[2] = w_tmp[3]

    # Check the output
    if np.isnan(w_out).any():
        raise DSPFatal("RC-CR^2 filter produced nans in output.")



from __future__ import annotations

import numpy as np
from numba import guvectorize

from ..utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    [
        "void(float32[:], float32, float32, float32, float32[:])",
        "void(float64[:], float64, float64, float64, float64[:])",
    ],
    "(n),(),(),()->(n)",
    **nb_kwargs,
)
def inject_gumbel(
    wf_in: np.ndarray, a: float, t0: float, beta: float, wf_out: np.ndarray
) -> None:
    """
    Injects a Gumbel distribution into the waveform `wf_in`, modifying it in place in `wf_out`.

    Parameters:
    - wf_in: Input waveform (1D array).
    - a: Amplitude of the Gumbel distribution.
    - t0: Temporal centroid of the Gumbel distribution.
    - beta: Scale parameter (controls spread of the Gumbel distribution).
    - wf_out: Output waveform (1D array), modified by adding the Gumbel distribution.
    """

    wf_out[:] = np.nan

    # Early exit if any of the inputs contain NaN values (invalid inputs).
    if np.isnan(wf_in).any() or np.isnan(a) or np.isnan(t0) or np.isnan(beta):
        return

    wf_out[:] = wf_in[:]

    # Define the range of indices over which the Gumbel distribution will be applied.
    # Start injecting the distribution at 2 beta below the centroid (t0).
    start = t0
    mu = t0 + (2 * beta)
    end = mu + (8 * beta)

    # Ensure the range is within valid waveform boundaries.
    if start < 0:
        start = 0
    if end > len(wf_in):
        end = len(wf_in)

    # Loop through the specified range and add the Gumbel distribution to wf_out.
    for i in range(start, end):

        z = (i - mu) / beta
        wf_out[i] += (a / beta) * np.exp(-(z + np.exp(-z)))


@guvectorize(
    [
        "void(float32[:], float32, float32, float32, float32, float32, float32, float32[:])",
        "void(float64[:], float64, float64, float64, float64, float64, float64, float64[:])",
    ],
    "(n),(),(),(),(),(),()->(n)",
    **nb_kwargs,
)
def inject_general_logistic(
    wf_in: np.ndarray,
    a: float,
    t0: float,
    rt: float,
    q: float,
    v: float,
    decay: float,
    wf_out: np.ndarray,
) -> None:
    r"""Inject sigmoid pulse into existing waveform to simulate pileup.

    .. math::

        s(t) = \frac{A}{(1 + q \exp[-4 \log(99) (t - t_0 - t_r/2) / t_r])^{1/v}}
                e^{-(t-t_0)/\tau}

    Parameters
    ----------
    wf_in
        the input waveform.
    a
        the amplitude :math:`A` of the injected waveform.
    t0
        the position :math:`t_0` of the injected waveform.
    rt
        the rise time :math:`t_r` of the injected waveform.
    q
        shaping parameter of the logistic function.
    v
        shaping parameter of the logistic function.
    decay
        the decay parameter :math:`\tau` of the injected waveform.
    wf_out
        the output waveform.
    """

    wf_out[:] = np.nan

    # Early exit if any of the inputs contain NaN values (invalid inputs).
    if (
        np.isnan(wf_in).any()
        or np.isnan(a)
        or np.isnan(t0)
        or np.isnan(rt)
        or np.isnan(q)
        or np.isnan(v)
        or np.isnan(decay)
    ):
        return

    wf_out[:] = wf_in[:]
    rise = 4 * np.log(99) / rt

    for t in range(len(wf_out)):
        wf_out[t] = wf_out[t] + a / (
            (1 + q * np.exp(-rise * (t - t0 - rt / 2))) ** (1 / v)
        ) * np.exp(-(1 / decay) * (t - t0))



import numpy as np
import pytest

from dspeed.errors import DSPFatal
from dspeed.processors import rc_cr2


def test_rc_cr2(compare_numba_vs_python):
    # Create a single exponential pulse to RC-CR^2 filter
    zeta = 30000
    w_len = 8192
    ts = np.arange(0, w_len, dtype=np.float64)
    amplitude = 17500
    tau = 500

    # ensure that if there is a nan in w_in, all nans are outputted
    w_in = np.ones(w_len)
    w_in[4] = np.nan

    assert np.all(np.isnan(compare_numba_vs_python(rc_cr2, w_in, tau)))

    # Ensure that if the input waveform is less than 3 samples, an error is raised
    with pytest.raises(DSPFatal):
        rc_cr2(np.zeros(3), tau, np.zeros(3))

    # ensure that a valid input gives the expected output
    pulse_in = np.zeros(w_len, dtype=np.float64)
    pulse_in[w_len // 2 :] = amplitude * np.exp(-ts[: w_len // 2] / zeta)
    out_pulse = np.zeros_like(pulse_in)
    rc_cr2(pulse_in, tau, out_pulse)

    t = np.arange(8192 // 2 + 1, dtype=np.float64)
    # This is the exact form of an RC-CR^2 filter applied to an exponential pulse...
    w_out_expected = np.zeros_like(pulse_in)
    w_out_expected[8192 // 2 - 1 :] = (
        -zeta * t**2 * np.exp(-t / tau) / (2 * tau * (zeta - tau))
        + t * (zeta**2 - 2 * zeta * tau) * np.exp(-t / tau) / (zeta - tau) ** 2
        + zeta * tau**3 * np.exp(-t / zeta) / (zeta - tau) ** 3
        - zeta * tau**3 * np.exp(-t / tau) / (zeta - tau) ** 3
    )
    w_out_expected *= amplitude
    # rescale because we abandoned gain scaling in the filter
    w_out_expected *= np.amax(out_pulse) / np.amax(w_out_expected)

    result = compare_numba_vs_python(rc_cr2, pulse_in, tau)
    assert result.dtype == np.float64
    assert np.allclose(
        compare_numba_vs_python(rc_cr2, pulse_in, tau), w_out_expected, rtol=1e-01
    )

    # Check that it works at float32 precision
    pulse_in = np.zeros(w_len, dtype=np.float32)
    pulse_in[w_len // 2 :] = amplitude * np.exp(-ts[: w_len // 2] / zeta)
    result = compare_numba_vs_python(rc_cr2, pulse_in, tau)
    assert result.dtype == np.float32
    assert np.allclose(result, w_out_expected, rtol=1e-01)
